"""Arguments needed to test inner geometry.

Each labware has 2 nominal volumes calculated in solidworks.
- One is a nominal bottom volume, calculated some set distance from the bottom of the inside of the well.
- The other is a nominal top volume, calculated some set distance from the top of the inside of the well.
"""

INNER_WELL_GEOMETRY_TEST_PARAMS = [
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell15mL",
        16.69,
        15546.930,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell50mL",
        111.2,
        56110.279,
        3.0,
        5.0,
    ],
    [
        "opentrons_24_tuberack_nest_2ml_screwcap",
        "conicalWell",
        64.4,
        2035.5,
        2.96,
        4.22,
    ],
    [
        "opentrons_24_tuberack_nest_1.5ml_screwcap",
        "conicalWell",
        18.8,
        1693.0,
        2.94,
        4.02,
    ],
    ["nest_1_reservoir_290ml", "cuboidalWell", 16570.12, 271690.52, 3.0, 3.0],
    [
        "opentrons_24_tuberack_nest_2ml_snapcap",
        "conicalWell",
        67.323,
        2077.6,
        2.946,
        4.06,
    ],
    ["nest_96_wellplate_2ml_deep", "cuboidalWell", 114.396, 1992.407, 2.93, 4.03],
    [
        "opentrons_24_tuberack_nest_1.5ml_snapcap",
        "conicalWell",
        27.8,
        1682.29,
        3.0,
        3.0,
    ],
    ["nest_12_reservoir_15ml", "cuboidalWell", 1178.77, 13236.1, 2.93, 3.0],
    ["nest_1_reservoir_195ml", "cuboidalWell", 14034.19, 172301.86, 3.0, 3.0],
    [
        "opentrons_24_tuberack_nest_0.5ml_screwcap",
        "conicalWell",
        21.226,
        769.19,
        3.15,
        3.35,
    ],
    [
        "opentrons_96_wellplate_200ul_pcr_full_skirt",
        "conicalWell",
        13.8,
        145.2,
        2.93,
        3.21,
    ],
    ["nest_96_wellplate_100ul_pcr_full_skirt", "conicalWell", 14.99, 145.79, 2.93, 3.2],
    ["nest_96_wellplate_200ul_flat", "conicalWell", 93.1, 251.2, 2.9, 3.24],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "50mlconicalWell",
        163.860,
        57720.510,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "15mlconicalWell",
        40.830,
        15956.6,
        3.0,
        3.0,
    ],
    ["usascientific_12_reservoir_22ml", "cuboidalWell", 66.17 * 8, 19797.38, 3.0, 6.0],
    [
        "thermoscientificnunc_96_wellplate_2000ul",
        "conicalWell",
        71.1,
        1768.02,
        2.94,
        3.0,
    ],
    ["usascientific_96_wellplate_2.4ml_deep", "cuboidalWell", 72.22, 2241.36, 3.0, 3.0],
    ["agilent_1_reservoir_290ml", "cuboidalWell", 15652.89, 268813.81, 3.0, 3.0],
    [
        "opentrons_24_tuberack_eppendorf_1.5ml_safelock_snapcap",
        "conicalWell",
        25.75,
        1576.14,
        3.0,
        3.0,
    ],
    [
        "thermoscientificnunc_96_wellplate_1300ul",
        "conicalWell",
        71.1,
        1117.0,
        2.94,
        3.71,
    ],
    ["corning_12_wellplate_6.9ml_flat", "conicalWell", 1118.1, 5468.2, 2.9, 3.46],
    ["corning_24_wellplate_3.4ml_flat", "conicalWell", 559.9, 2759.3, 2.9, 3.46],
    ["corning_6_wellplate_16.8ml_flat", "conicalWell", 2767.7, 13443.1, 2.9, 3.47],
    ["corning_48_wellplate_1.6ml_flat", "conicalWell", 260.0, 1283.2, 2.9, 3.46],
    ["biorad_96_wellplate_200ul_pcr", "conicalWell", 17.3, 155.9, 2.93, 3.23],
    ["axygen_1_reservoir_90ml", "cuboidalWell", 14909.08, 72854.8, 2.0, 2.68],
    ["corning_384_wellplate_112ul_flat", "flatWell", 22.43, 77.39, 2.88, 3.0],
    ["corning_96_wellplate_360ul_flat", "conicalWell", 94.0, 248.6, 2.9, 3.24],
    ["biorad_384_wellplate_50ul", "conicalWell", 7.7, 27.76, 3.0, 3.0],
    [
        "appliedbiosystemsmicroamp_384_wellplate_40ul",
        "conicalWell",
        7.2,
        25.3,
        2.94,
        3.11,
    ],
    [
        "opentrons_24_tuberack_eppendorf_2ml_safelock_snapcap",
        "conicalWell",
        60.94,
        2163.98,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell15mL",
        16.69,
        15546.93,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical",
        "conicalWell50mL",
        111.2,
        56110.279,
        3.0,
        5.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "15mlconicalWell",
        40.83,
        15956.6,
        3.0,
        3.0,
    ],
    [
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical",
        "50mlconicalWell",
        163.86,
        57720.51,
        3.0,
        3.0,
    ],
]
