"""Integration tests for the Python Protocol API.

These test the Python Protocol API from the point of view of the user's Python protocol.

They do not make sure the robot would actually move to the right place--that is the job of
other tests, such as g-code-testing.

These are supplementary for Python Protocol API features that depend on nontrivial interaction
between those layers.
"""
